from __future__ import annotations
from __future__ import absolute_import
from __future__ import print_function
import csv
from datetime import datetime
from .._common import BaseServer
from arcgis.gis import GIS
from typing import Optional


########################################################################
class LogManager(BaseServer):
    """
    Helper class for the management of logs by administrators.

    Logs are the transaction records written by the various components
    of ArcGIS Server.  You can query the logs, change various log settings,
    and check error messages for helping to determine the nature of an issue.

    """

    _url = None
    _con = None
    _json_dict = None
    _json = None

    # ----------------------------------------------------------------------
    def __init__(self, url: str, gis: GIS, initialize: bool = False):
        """Constructor


        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        url                    Required string. The machine URL.
        ------------------     --------------------------------------------------------------------
        gis                    Optional string. The GIS or Server object..
        ==================     ====================================================================

        """
        connection = gis
        super(LogManager, self).__init__(gis=gis, url=url)
        self._url = url
        self._con = connection
        if initialize:
            self._init(connection)

    # ----------------------------------------------------------------------
    def __str__(self) -> str:
        return "<%s at %s>" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    def __repr__(self) -> str:
        return "<%s at %s>" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    def count_error_reports(self, machine: str = "*") -> dict:
        """
        This operation counts the number of error reports (crash reports) that have been generated
        on each machine.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        machine                Optional string. The name of the machine on which to count the
                               reports. The default will return the count for all machines in a site.
        ==================     ====================================================================

        :return:
           A dictionary with machine name and report count.

        """
        params = {"f": "json", "machine": machine}
        url = self._url + "/countErrorReports"
        return self._con.post(path=url, postdata=params)

    # ----------------------------------------------------------------------
    def clean(self) -> bool:
        """
        Deletes all the log files on all server machines in the site. This is an irreversible
        operation.

        This operation forces the server to clean the logs, which has the effect of freeing
        up disk space. However, it is not required that you invoke this operation because
        the server periodically purges old logs.

        :return:
           A boolean indicating success (True) or failure (False).

        """
        params = {
            "f": "json",
        }
        url = "{}/clean".format(self._url)
        res = self._con.post(path=url, postdata=params)
        if "status" in res:
            return res["status"] == "success"
        return res

    # ----------------------------------------------------------------------
    @property
    def settings(self) -> dict:
        """Gets the current log settings."""
        params = {"f": "json"}
        url = self._url + "/settings"
        try:
            return self._con.post(path=url, postdata=params)["settings"]
        except:
            return ""

    # ----------------------------------------------------------------------
    def edit(
        self,
        level: str = "WARNING",
        log_dir: Optional[str] = None,
        max_age: int = 90,
        max_report_count: int = 10,
    ) -> dict:
        """
        Provides log editing capabilities for the entire site.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        level                  Optional string. The log level.  Can be one of (in severity order):
                               OFF, DEBUG, VERBOSE, FINE, INFO, WARNING, SEVERE. The default is WARNING.
        ------------------     --------------------------------------------------------------------
        log_dir                Optional string. The file path to the root of the log directory.
        ------------------     --------------------------------------------------------------------
        max_age                Optional integer. The number of days that a server should save a
                               log file. The default is 90.
        ------------------     --------------------------------------------------------------------
        ax_report_count        Optional integer. The maximum number of error report files per
                               machine. The default is 10.
        ==================     ====================================================================


        :return:
           A JSON with the edited settings.

        """
        url = self._url + "/settings/edit"
        allowed_levels = (
            "OFF",
            "SEVERE",
            "WARNING",
            "INFO",
            "FINE",
            "VERBOSE",
            "DEBUG",
        )
        current_settings = self.settings
        current_settings["f"] = "json"

        if level.upper() in allowed_levels:
            current_settings["logLevel"] = level.upper()
        if log_dir is not None:
            current_settings["logDir"] = log_dir
        if max_age is not None and isinstance(max_age, int):
            current_settings["maxLogFileAge"] = max_age
        if (
            max_report_count is not None
            and isinstance(max_report_count, int)
            and max_report_count > 0
        ):
            current_settings["maxErrorReportsCount"] = max_report_count
        return self._con.post(path=url, postdata=current_settings)

    # ----------------------------------------------------------------------
    def query(
        self,
        start_time: int | datetime | None = None,
        end_time: int | datetime | None = None,
        since_server_start: bool = False,
        level: str = "WARNING",
        services: str = "*",
        machines: str = "*",
        server: str = "*",
        codes: str | None = None,
        process_IDs: str | None = None,
        export: bool = False,
        export_type: str = "CSV",
        out_path: str | None = None,
        max_records_return: int = 5000,
    ):
        """
        The query operation on the logs resource provides a way to
        aggregate, filter, and page through logs across the entire site.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        start_time             Optional Integer or datetime. The most recent time to query.  Default is now.
                               Time can be specified in milliseconds since UNIX epoch.


                               Example for integer:
                               start_time = 1312237040123

        ------------------     --------------------------------------------------------------------
        end_time               Optional String or datetime. The oldest time to include in the result set. You
                               can use this to limit the query to the last n minutes or hours as
                               needed.

                               If ```sinceLastStart``` is true, the default is all logs since the
                               server was started.
        ------------------     --------------------------------------------------------------------
        since_server_start     Optional Bool. Gets only the records written since the server
                               started (True).  The default is False.
        ------------------     --------------------------------------------------------------------
        level                  Optional String. Gets only the records with a log level at or more
                               severe than the level declared here. Can be one of (in severity
                               order): DEBUG, VERBOSE, FINE, INFO, WARNING, SEVERE. The
                               default is WARNING.
        ------------------     --------------------------------------------------------------------
        services               Optional String. Query records related to a specific service.
                               The default is all.
        ------------------     --------------------------------------------------------------------
        machines               Optional String. Query records related to a specific machine.
                               The default is all.
        ------------------     --------------------------------------------------------------------
        server                 Optional String. Query records related to a specific server.
                               The default is all.
        ------------------     --------------------------------------------------------------------
        codes                  Optional String. Gets only the records with the specified code.
                               The default is all.  See http://server.arcgis.com/en/server/latest/administer/windows/log-codes-overview.htm
        ------------------     --------------------------------------------------------------------
        process_IDs            Optional String. Query by the machine process ID that logged the event.
        ------------------     --------------------------------------------------------------------
        export                 Optional String. Boolean indicating whether to export the query
                               results.  The default is False (don't export).
        ------------------     --------------------------------------------------------------------
        export_type            Optional String. The export file type. CSV or TAB are the choices,
                               CSV is the default.
        ------------------     --------------------------------------------------------------------
        out_path               Optional String. The path to download the log file to.
        ------------------     --------------------------------------------------------------------
        max_records_return     Optional Int. The maximum amount of records to return. Default is 5000
        ==================     ====================================================================

        :return:
           A JSON of the log items that match the query. If export option is set to True, the
           output log file path is returned.


        """

        if codes is None:
            codes = []
        if process_IDs is None:
            process_IDs = []
        allowed_levels = ("SEVERE", "WARNING", "INFO", "FINE", "VERBOSE", "DEBUG")
        qFilter = {"services": "*", "machines": "*", "server": "*"}
        if len(process_IDs) > 0:
            qFilter["processIds"] = process_IDs
        if len(codes) > 0:
            qFilter["codes"] = codes
        params = {
            "f": "json",
            "sinceServerStart": since_server_start,
        }
        # set page size
        params["pageSize"] = max_records_return if max_records_return < 5000 else 5000
        # if greater than 5000 then will enter hasMore loop else, correct number returned
        max_records_return -= 5000
        url = "{url}/query".format(url=self._url)
        if start_time is not None and isinstance(start_time, datetime):
            params["startTime"] = int(start_time.timestamp() * 1000)
        elif start_time:
            params["startTime"] = start_time
        else:
            params["startTime"] = int(datetime.now().timestamp() * 1000)
        if end_time is not None and isinstance(end_time, datetime):
            params["endTime"] = int(end_time.timestamp() * 1000)
        elif end_time and isinstance(end_time, int):
            params["endTime"] = end_time
        if level.upper() in allowed_levels:
            params["level"] = level
        if server != "*":
            qFilter["server"] = server.split(",")
        if services != "*":
            qFilter["services"] = services.split(",")
        if machines != "*":
            qFilter["machines"] = machines.split(",")
        params["filter"] = qFilter

        logs = self._con.post(path=url, postdata=params)
        # determine if more logs are available to query
        has_more = logs["hasMore"]

        # If the hasMore member of the response object is true,
        # pass the end time as the startTime parameter
        # for the next request to get the next set of records
        loop = 0
        new_logs = {}
        while max_records_return > 1 and has_more:
            if has_more:
                # get new start time from logs endTime in first loop then from new_logs endTime after
                params["startTime"] = (
                    logs["endTime"] if loop == 0 else new_logs["endTime"]
                )
                loop = 1
                # page size is 5000 or less
                params["pageSize"] = (
                    max_records_return if max_records_return < 5000 else 5000
                )
                # decides if more records needed or not
                max_records_return -= 5000
                # new logs to query
                new_logs = self._con.post(path=url, postdata=params)
                has_more = new_logs["hasMore"]
                # append new log messages to logs to return
                for log_message in new_logs["logMessages"]:
                    logs["logMessages"].append(log_message)
                else:
                    break
        # if export true then no values returned, file written to
        if export is True and out_path is not None:
            with open(file=out_path, mode="w") as f:
                hasKeys = False
                if export_type == "TAB":
                    csvwriter = csv.writer(f, delimiter="\t")
                else:
                    csvwriter = csv.writer(f)
                for message in logs["logMessages"]:
                    if hasKeys == False:
                        csvwriter.writerow(list(message.keys()))
                        hasKeys = True
                    csvwriter.writerow(list(message.values()))
                    del message
            del logs
            return out_path
        else:
            return logs



from typing import Optional
from arcgis._impl.common._mixins import PropertyMap


###########################################################################
class CertificateManager(object):
    """
    The ``CertificateManager`` class provides the administrator the ability to
    register and unregister certificates with the :class:`~arcgis.gis.GIS`.

    .. note::
        This resource is
        available via HTTPS only.
    """

    _gis = None
    _con = None
    _url = None
    _properties = None

    def __init__(self, gis):
        self._url = gis._portal.resturl + "portals/self/certificates"
        self._gis = gis
        self._con = gis._con

    # ----------------------------------------------------------------------
    def _init(self):
        result = self._con.get(self._url, {"f": "json"})
        self._properties = PropertyMap(result)

    # ----------------------------------------------------------------------
    @property
    def properties(self):
        """
        The ``properties`` method retrieves the properties of the certificate

        :return:
            A list of the certificate properties
        """
        self._init()
        return self._properties

    # ----------------------------------------------------------------------
    def add(self, name: str, domain: str, certificate: str):
        """
        The ``add`` method allows allows administrators to
        register custom X.509 HTTPS certificates with their ArcGIS Online
        organizations. This will allow ArcGIS Online organization to trust
        the custom certificates used by a remote server when making HTTPS
        requests to it, i.e. store credentials required to access its
        resource and share with others.

        .. note::
            A maximum of 5 certificates can be registered with an organization.

        ================  ===============================================================================
        **Parameter**     **Description**
        ----------------  -------------------------------------------------------------------------------
        name              Required String. The certificate name.
        ----------------  -------------------------------------------------------------------------------
        domain            Required String. Server domain that the certificate is used for.
        ----------------  -------------------------------------------------------------------------------
        certificate	      Required String. Base64-encoded certificate text, enclosed between
                          `-----BEGIN CERTIFICATE-----` and `-----END CERTIFICATE-----`.
        ================  ===============================================================================

        :return:
            A boolean indicating success (True), or failure (False)

        .. code-block:: python

            # Usage Example

            >>> gis.admin.certificates.add(name="certificate_name",
            >>>                            domain = "domain_name",
            >>>                            certificate = "certificate_text")

        """
        url = self._url + "/register"
        params = {
            "f": "json",
            "name": name,
            "domain": domain,
            "certificate": certificate,
        }
        import json

        res = self._con.post(url, params, try_json=False)
        res = res.replace(",}", "}")
        res = json.loads(res)
        if "success" in res:
            return res["success"]
        return res

    # ----------------------------------------------------------------------
    def get(self, cert_id: str):
        """
        The ``get`` method retrieves the certificate information for a single certificate

        ================  ===============================================================================
        **Parameter**     **Description**
        ----------------  -------------------------------------------------------------------------------
        cert_id           Required String.  The ID of the certificate to delete.
        ================  ===============================================================================

        .. code-block:: python

            # Usage Example

            >>> gis.admin.certificates.get(cert_id= "certificate_id")

        :return:
            A Dictionary (if found), else None

        The dictionary contains the following information:

        ================  ===============================================================================
        **Key**           **Value**
        ----------------  -------------------------------------------------------------------------------
        id                The ID of the registered certificate.
        ----------------  -------------------------------------------------------------------------------
        name              The certificate name.
        ----------------  -------------------------------------------------------------------------------
        domain            Server domain that the certificate is used for.
        ----------------  -------------------------------------------------------------------------------
        sslCertificate	  Base64-encoded certificate text, enclosed between `-----BEGIN CERTIFICATE-----` and `-----END CERTIFICATE-----`.
        ================  ===============================================================================

        """
        found_cert_id = None
        for cert in self.certificates:
            if cert_id.lower() == cert["id"].lower():
                found_cert_id = cert["id"]
                break
        if found_cert_id:
            url = self._url + "/{found_cert_id}".format(found_cert_id=found_cert_id)
            return self._con.get(url, {"f": "json"})
        return None

    # ----------------------------------------------------------------------
    def delete(self, cert_id: str):
        """
        The ``delete`` method unregisters the certificate from the organization.

        ================  ===============================================================================
        **Parameter**     **Description**
        ----------------  -------------------------------------------------------------------------------
        cert_id           Required String.  The ID of the certificate to delete.
        ================  ===============================================================================

        :return:
            A boolean indicating success (True), or failure (False)

        .. code-block:: python

            # Usage Example

            >>> gis.admin.certificates.delete(cert_id="certificate_id")

        """
        url = self._url + "/{cert_id}/unregister".format(cert_id=cert_id)
        params = {"f": "json"}
        res = self._con.post(url, params)
        if "success" in res:
            return res["success"]
        return res

    # ----------------------------------------------------------------------
    def update(
        self,
        cert_id: str,
        name: Optional[str] = None,
        domain: Optional[str] = None,
        certificate: Optional[str] = None,
    ):
        """
        The ``update`` operation allows organization's
        administrators to update a registered custom X.509 HTTPS
        certificate.

        ================  ===============================================================================
        **Parameter**     **Description**
        ----------------  -------------------------------------------------------------------------------
        cert_id           Required String.  The ID of the certificate to delete.
        ----------------  -------------------------------------------------------------------------------
        name              Optional String. The certificate name.
        ----------------  -------------------------------------------------------------------------------
        domain            Optional String. Server domain that the certificate is used for.
        ----------------  -------------------------------------------------------------------------------
        certificate	      Optional String. Base64-encoded certificate text, enclosed between `
                          -----BEGIN CERTIFICATE-----` and `-----END CERTIFICATE-----`.
        ================  ===============================================================================

        :return:
            A boolean indicating success (True), or failure (False)

        .. code-block:: python

            # Usage Example

            >>> gis.admin.certificates.update(cert_id ="certificate_id",
            >>>                               name = "certificate_name",
            >>>                               domain ="certificate_domain",
            >>>                               certificate = "certificate_text")

        """
        url = self._url + "/{cert_id}/update".format(cert_id=cert_id)
        params = {"f": "json"}

        if not name is None:
            params["name"] = name
        if not domain is None:
            params["domain"] = domain
        if not certificate is None:
            params["certificate"] = certificate
        import json  # HANDLES BUG IN API

        res = self._con.post(url, params, try_json=False)
        res = json.loads(res.replace(",}", "}"))
        if "success" in res:
            return res["success"]
        return res

    # ----------------------------------------------------------------------
    @property
    def certificates(self):
        """
        The ``certificates`` property retrieves the list of certificates registered with the organization

        :return:
            A List containing the information of registered certificates
        """
        return self.properties.certificates



from ._codetemplate import super_resolution
import torch, json, traceback
from ... import __version__ as ArcGISLearnVersion
from .._data import prepare_data, _raise_fastai_import_error

try:
    from ._arcgis_model import ArcGISModel, _resnet_family, _EmptyData, _get_device
    from ._superres_utils import (
        compute_metrics,
        create_loss,
        UNetSR,
    )
    from ._SR3_utils import (
        UNet,
        GaussianDiffusion,
        init_weights,
        l1Loss,
    )
    from fastai.vision.learner import unet_learner, cnn_config
    from fastai.vision import nn, NormType, Learner, optim
    from fastai.callbacks import LossMetrics
    from fastai.utils.mem import Path
    from .._utils.common import _get_emd_path, ArcGISMSImage
    from .._utils.env import is_arcgispronotebook
    from fastai.core import ifnone
    from .._utils.superres import show_results
    from .._data_utils.pix2pix_data import normalize, denormalize, prepare_pix2pix_data

    HAS_FASTAI = True
except Exception as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_FASTAI = False


class SuperResolution(ArcGISModel):

    """
    Creates a model object which increases the resolution and improves the quality of images.
    Based on Fast.ai MOOC Lesson 7 and https://github.com/Janspiry/Image-Super-Resolution-via-Iterative-Refinement.

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            :meth:`~arcgis.learn.prepare_data` function.
    ---------------------   -------------------------------------------
    backbone                Optional string. Backbone CNN model to be used for
                            creating the base of the
                            :class:`~arcgis.learn.SuperResolution`, which
                            is `resnet34` by default.
                            Compatible backbones: 'SR3', 'resnet18', 'resnet34',
                            'resnet50', 'resnet101', 'resnet152'.
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained model is
                            saved.
    =====================   ===========================================

    In addition to explicitly named parameters, the SuperResolution model with 'SR3' backbone supports the optional key word arguments:

    **kwargs**

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    inner_channel           Optional int. Channel dimension.
                            Default: 64.
    ---------------------   -------------------------------------------
    norm_groups             Optional int. Group normalization.
                            Default: 32
    ---------------------   -------------------------------------------
    channel_mults           Optional int. Depth or channel multipliers.
                            Default: [1, 2, 4, 4, 8, 8]
    ---------------------   -------------------------------------------
    attn_res                Optional int. Number of attention in residual blocks.
                            Default: 16
    ---------------------   -------------------------------------------
    res_blocks              Optional int. Number of resnet block.
                            Default: 3
    ---------------------   -------------------------------------------
    dropout                 Optional bool. Dropout.
                            Default: 0
    ---------------------   -------------------------------------------
    schedule                Optional int. Type of noise schedule. available types
                            are "linear", 'warmup10', 'warmup50', 'const', 'jsd',
                            'cosine'. Default: 'linear'
    ---------------------   -------------------------------------------
    n_timestep              Optional int. Number of time-steps.
                            Default: 1000
    ---------------------   -------------------------------------------
    linear_start            Optional bool. Schedule start.
                            Default: 1e-06
    ---------------------   -------------------------------------------
    linear_end              Optional bool. Schedule end.
                            Default: 1e-02
    =====================   ===========================================

    :return: :class:`~arcgis.learn.SuperResolution` Object
    """

    def __init__(self, data, backbone=None, pretrained_path=None, *args, **kwargs):
        self._learn_version = kwargs.get("ArcGISLearnVersion", ArcGISLearnVersion)
        if backbone == "SR3":
            data_bunch = None
            if data.train_ds.__class__.__name__ == "Pix2PixHDDataset":
                downsampling_factor = data._downsampling_factor
                val_split_pct = data.val_split_pct
            else:
                downsampling_factor = data.downsample_factor
                val_split_pct = data._val_split_pct

            if isinstance(data.train_ds, list):
                data_bunch = data
            else:
                data_bunch = prepare_data(
                    path=data.train_ds.path,
                    batch_size=data.batch_size,
                    downsample_factor=downsampling_factor,
                    val_split_pct=val_split_pct,
                    seed=data.seed,
                    dataset_type="SR3",
                )
            super().__init__(data_bunch, backbone, **kwargs)
            self.kwargs = kwargs
            self._data = data_bunch if data_bunch else data
            denoiseUnet = UNet(
                in_channel=(self._data._n_channel) * 2,
                out_channel=self._data._n_channel,
                image_size=self._data.chip_size,
                with_noise_level_emb=True,
                **kwargs
            )
            sr3model = GaussianDiffusion(
                denoiseUnet,
                image_size=self._data.chip_size,
                channels=self._data._n_channel,
                device=self._device.type,
            )
            init_weights(sr3model, init_type="orthogonal")
            sr3model.set_new_noise_schedule(self._device.type, **kwargs)
            self.learn = Learner(
                self._data,
                sr3model,
                loss_func=l1Loss(self._device.type),
                opt_func=optim.Adam,
            )
            self.model_type = "SR3"
        else:
            data_bunch = None
            if data.train_ds.__class__.__name__ == "Pix2PixHDDataset":
                data_bunch = prepare_data(
                    path=data.path,
                    batch_size=data.batch_size,
                    downsample_factor=data._downsampling_factor,
                    val_split_pct=data.val_split_pct,
                    seed=data.seed,
                    dataset_type="superres",
                )
            super().__init__(data, backbone, **kwargs)
            self._data = data_bunch if data_bunch else data
            self._data._extract_bands = list(range(self._data._n_channel))
            feat_loss = create_loss(self._data._n_channel, self._device.type)
            self._check_dataset_support(self._data)
            if self._data._is_multispec:
                model = UNetSR(
                    self._data,
                    arch=self._backbone,
                    norm_type=NormType.Weight,
                )
                self.learn = Learner(
                    self._data,
                    model,
                    wd=1e-3,
                    loss_func=feat_loss,
                    callback_fns=LossMetrics,
                )
                self.learn.split(ifnone(None, cnn_config(self._backbone)["split"]))
            else:
                attention = True if self._learn_version > "2.1.0.3" else False
                self._data.c = self._data._n_channel
                self.learn = unet_learner(
                    self._data,
                    arch=self._backbone,
                    wd=1e-3,
                    loss_func=feat_loss,
                    callback_fns=LossMetrics,
                    blur=True,
                    self_attention=attention,
                    norm_type=NormType.Weight,
                )
            self.model_type = "UNet"
        self.learn.data = self._data
        self.learn.model = self.learn.model.to(self._device)
        if pretrained_path is not None:
            self.load(pretrained_path)

        self._code = super_resolution

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<%s>" % (type(self).__name__)

    @staticmethod
    def _available_metrics():
        return ["valid_loss"]

    @property
    def supported_backbones(self):
        """
        Supported torchvision backbones for this model.
        """
        return SuperResolution._supported_backbones()

    @staticmethod
    def _supported_backbones():
        return ["SR3", *_resnet_family]

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a :class:`~arcgis.learn.SuperResolution` object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from :meth:`~arcgis.learn.prepare_data` function or None for
                                inferencing.
        =====================   ===========================================

        :return: :class:`~arcgis.learn.SuperResolution` Object
        """
        return cls.from_emd(data, emd_path)

    @classmethod
    def from_emd(cls, data, emd_path):
        """
        Creates a SuperResolution object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from :meth:`~arcgis.learn.prepare_data` function or None for
                                inferencing.
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Esri Model Definition
                                file.
        =====================   ===========================================

        :return: :class:`~arcgis.learn.SuperResolution` Object
        """

        if not HAS_FASTAI:
            _raise_fastai_import_error(import_exception=import_exception)
        emd_path = _get_emd_path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)
        model_file = Path(emd["ModelFile"])
        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file
        modtype = emd.get("ModelArch", "UNet")
        model_params = emd["ModelParameters"]
        downsample_factor = emd.get("downsample_factor")
        n_channel = emd.get("n_channel", 3)
        resize_to = emd.get("resize_to")
        chip_size = emd["ImageHeight"]
        kwargs = emd.get("Kwargs", {})
        kwargs["ArcGISLearnVersion"] = emd.get("ArcGISLearnVersion", "1.0.0")

        if data is None:
            if modtype == "SR3":
                data = _EmptyData(
                    path=emd_path.parent, loss_func=None, c=2, chip_size=chip_size
                )
                data._val_split_pct = 0.1
                data._image_stats = emd.get("image_stats")
                data._image_stats2 = emd.get("image_stats2", None)
            else:
                if emd.get("is_multispec", False):
                    data = _EmptyData(
                        path=emd_path.parent, loss_func=None, c=2, chip_size=chip_size
                    )
                    data._train_tail = False
                    data._image_stats = emd.get("image_stats")
                    data._image_stats2 = emd.get("image_stats2", None)
                else:
                    data = _EmptyData(
                        path=emd_path.parent, loss_func=None, c=2, chip_size=chip_size
                    )
            data._is_multispec = emd.get("is_multispec", False)
            data._n_channel = n_channel
            data._is_empty = True
            data.emd_path = emd_path
            data.downsample_factor = downsample_factor
            data.emd = emd
            data._extract_bands = emd.get("extract_bands", None)
            data._bands = emd.get("bands", None)
            data.device = _get_device()
        backbone = "SR3" if modtype == "SR3" else model_params.get("backbone")
        data.resize_to = resize_to

        return cls(data, backbone=backbone, pretrained_path=str(model_file), **kwargs)

    @property
    def _model_metrics(self):
        return self.compute_metrics(show_progress=True)

    def _get_emd_params(self, save_inference_file):
        _emd_template = {}
        _emd_template["Framework"] = "arcgis.learn.models._inferencing"
        _emd_template["ModelConfiguration"] = "_superres"
        if save_inference_file:
            _emd_template["InferenceFunction"] = "ArcGISSuperResolution.py"
        else:
            _emd_template[
                "InferenceFunction"
            ] = "[Functions]System\\DeepLearning\\ArcGISLearn\\ArcGISSuperResolution.py"
        _emd_template["downsample_factor"] = self._data.downsample_factor
        _emd_template["n_channel"] = self._data._n_channel
        _emd_template["is_multispec"] = self._data._is_multispec
        _emd_template["ModelType"] = "SuperResolution"

        if self._data.train_ds.__class__.__name__ == "SR3Dataset":
            _emd_template["ModelArch"] = "SR3"
            _emd_template["image_stats"] = {
                i: j.tolist() for i, j in self._data.batch_stats_a.items() if j != None
            }
            _emd_template["image_stats2"] = {
                i: j.tolist() for i, j in self._data.batch_stats_b.items() if j != None
            }
            _emd_template["Kwargs"] = self.kwargs
        else:
            _emd_template["ModelArch"] = "UNet"
            _emd_template["image_stats"] = self._data._image_stats
            _emd_template["image_stats2"] = self._data._image_stats2
            _emd_template["extract_bands"] = self._data._extract_bands
            _emd_template["bands"] = self._data._bands
        return _emd_template

    def compute_metrics(self, accuracy=True, show_progress=True, **kwargs):
        """
        Computes Peak Signal-to-Noise Ratio (PSNR) and
        Structural Similarity Index Measure (SSIM) on validation set.

        """
        self._check_requisites()
        psnr, ssim = compute_metrics(self, self._data.valid_dl, show_progress, **kwargs)
        return {"PSNR": "{0:1.4e}".format(psnr), "SSIM": "{0:1.4e}".format(ssim)}

    def show_results(self, rows=None, **kwargs):
        """
        Displays the results of a trained model on a part of the validation set.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional int. Number of rows of results
                                to be displayed.
        =====================   ===========================================

        **kwargs**

        =====================   ===========================================
        sampling_type           Optional string. Type of sampling.
                                Default: 'ddim'. keyword arguments applicable for
                                SR3 model type only.
        ---------------------   -------------------------------------------
        n_timestep              Optional int. Number of time-steps for the sampling process.
                                Default: 200
        =====================   ===========================================

        """
        if not rows:
            if self.model_type == "UNet":
                rows = 5
            else:
                rows = 1
        if rows > len(self._data.valid_ds):
            rows = len(self._data.valid_ds)

        self._check_requisites()
        show_results(self, rows, **kwargs)
        if is_arcgispronotebook():
            from matplotlib import pyplot as plt

            plt.show()

    def predict(self, img_path):
        """
        Predicts and display the image.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        img_path                Required path of an image.
        =====================   ===========================================

        """
        from ..models._inferencing.util import mean, std

        img_path = Path(img_path)
        raw_img = ArcGISMSImage.open(img_path)
        raw_img = raw_img.resize(self._data.chip_size)
        d_mean, d_std = mean, std

        if self._data._is_multispec:
            mean, std = self._data._image_stats[0], self._data._image_stats[1]
            d_mean, d_std = (
                self._data._image_stats2[0],
                self._data._image_stats2[1],
            )

        raw_img_tensor = normalize(raw_img.px, mean, std)
        raw_img_tensor = raw_img_tensor[None].to(self._device)

        self.learn.model.eval()
        with torch.no_grad():
            prediction = self.learn.model(raw_img_tensor)[0].detach()[0].cpu()

        pred_denorm = ArcGISMSImage(denormalize(prediction, d_mean, d_std))
        # pred_denorm = pred_denorm.show()
        return pred_denorm

    @property
    def supported_datasets(self):
        """Supported dataset types for this model."""
        return SuperResolution._supported_datasets()

    @staticmethod
    def _supported_datasets():
        return ["Export_Tiles", "superres"]



import traceback
from .._data import _raise_fastai_import_error
from ._inference_only_models import InferenceOnlyModel

HAS_TRANSFORMER = True

try:
    from pathlib import Path
    import json
    import os
    import torch
    from transformers import pipeline, logging
    from fastprogress.fastprogress import progress_bar
    from transformers import AutoTokenizer, AutoModelForSeq2SeqLM
except Exception as e:
    transformer_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_TRANSFORMER = False


class TextTranslator(InferenceOnlyModel):
    """
    Creates a :class:`~arcgis.learn.text.TextTranslator` Object.
    Based on the Hugging Face transformers library
    To learn more about the available models for translation task,
    kindly visit:- https://huggingface.co/models?pipeline_tag=translation&search=Helsinki

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    source_language         Optional string. Specify the language of the
                            text you would like to get the translation of.
                            Default value is 'es' (Spanish)
    ---------------------   -------------------------------------------
    target_language         Optional string. The language into which one
                            wishes to translate the input text.
                            Default value is 'en' (English)
    =====================   ===========================================

    **kwargs**

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    pretrained_path         Option str. Path to a directory, where pretrained
                            model files are saved.
                            If pretrained_path is provided, the model is
                            loaded from that path on the local disk.
    ---------------------   -------------------------------------------
    working_dir             Option str. Path to a directory on local filesystem.
                            If directory is not present, it will be created.
                            This directory is used as the location to save the
                            model.
    =====================   ===========================================

    :return: :class:`~arcgis.learn.text.TextTranslator` Object
    """

    #: supported transformer architectures
    supported_backbones = ["MarianMT"]

    def __init__(self, source_language="es", target_language="en", **kwargs):
        self._source_lang = source_language
        self._target_lang = target_language
        self._task = f"translation_{self._source_lang}_to_{self._target_lang}"
        if not HAS_TRANSFORMER:
            _raise_fastai_import_error(import_exception=transformer_exception)
        super().__init__(task=self._task, **kwargs)

    def _load_model(self):
        if self._pretrained_path:
            pretrained_path = r"{}".format(self._pretrained_path)
            self._tokenizer = AutoTokenizer.from_pretrained(pretrained_path)
            self.model = AutoModelForSeq2SeqLM.from_pretrained(pretrained_path)
        else:
            self._tokenizer = AutoTokenizer.from_pretrained(
                f"Helsinki-NLP/opus-mt-{self._source_lang}-{self._target_lang}"
            )
            self.model = AutoModelForSeq2SeqLM.from_pretrained(
                f"Helsinki-NLP/opus-mt-{self._source_lang}-{self._target_lang}"
            )

        device = torch.device(
            "cpu" if self._device < 0 else "cuda:{}".format(self._device)
        )
        self.model.to(device)

    def translate(self, text_or_list, show_progress=True, **kwargs):
        """
        Translate the given text or list of text into the target language

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        text_or_list            Required string or list. A text/passage
                                or a list of texts/passages to translate.
        ---------------------   -------------------------------------------
        show_progress           optional Bool. If set to True, will display a
                                progress bar depicting the items processed so far.
        =====================   ===========================================

        **kwargs**

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        min_length              Optional integer. The minimum length of the
                                sequence to be generated.
                                Default value is set to to `min_length` parameter
                                of the model config.
        ---------------------   -------------------------------------------
        max_length              Optional integer. The maximum length of the
                                sequence to be generated.
                                Default value is set to to `max_length` parameter
                                of the model config.
        ---------------------   -------------------------------------------
        num_return_sequences    Optional integer. The number of independently
                                computed returned sequences for each element
                                in the batch.
                                Default value is set to 1.
        ---------------------   -------------------------------------------
        num_beams               Optional integer. Number of beams for beam
                                search. 1 means no beam search.
                                Default value is set to 1.
        ---------------------   -------------------------------------------
        length_penalty          Optional float. Exponential penalty to the
                                length. 1.0 means no penalty. Set to values < 1.0
                                in order to encourage the model to generate
                                shorter sequences, to a value > 1.0 in order to
                                encourage the model to produce longer sequences.
                                Default value is set to 1.0.
        ---------------------   -------------------------------------------
        early_stopping          Optional bool. Whether to stop the beam search
                                when at least ``num_beams`` sentences are
                                finished per batch or not.
                                Default value is set to False.
        =====================   ===========================================

        :return: a list or a list of list containing the translation of the input prompt(s) / sentence(s) to the target language
        """
        results = []
        num_return_sequences = kwargs.get("num_return_sequences", 1)
        min_length, max_length = kwargs.get("min_length"), kwargs.get("max_length")
        if min_length and max_length and min_length > max_length:
            error_message = (
                f"Value of `min_length` parameter({min_length}) cannot be "
                f"greater than the value of `max_length` parameter({max_length})."
            )
            raise Exception(error_message)

        if not isinstance(text_or_list, (list, tuple)):
            text_or_list = [text_or_list]

        for i in progress_bar(range(len(text_or_list)), display=show_progress):
            inputs = self._tokenizer.encode(
                f"{text_or_list[i]} {self._tokenizer.eos_token}", return_tensors="pt"
            ).to(self._device)
            outputs = self.model.generate(inputs, **kwargs)
            result = self._tokenizer.batch_decode(outputs, skip_special_tokens=True)
            result = [{"translated_text": x} for x in result]
            if num_return_sequences == 1:
                result = result[0]
            results.append(result)
        return results

    def _create_emd(self, name_or_path):
        emd_template = {}
        emd_template.update({"ModelName": self.__class__.__name__})
        emd_template.update({"architectures": self.model.config.architectures})
        emd_template.update({"source_lang": self._source_lang})
        emd_template.update({"target_lang": self._target_lang})
        path = Path(name_or_path)
        name = path.parts[-1]
        with open(os.path.join(path, f"{name}.emd"), "w") as f:
            f.write(json.dumps(emd_template))

    def save(self, name_or_path):
        """
        Saves the translator model files on a specified path on the local disk.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Path to save
                                model files on the local disk.
        =====================   ===========================================

        :return: Absolute path for the saved model
        """
        if "\\" in name_or_path or "/" in name_or_path:
            path = name_or_path
        else:
            path = os.path.join(self.working_dir, "models", name_or_path)
        self.model.save_pretrained(path)
        self.model.config.save_pretrained(path)
        self._tokenizer.save_pretrained(path)
        self._create_emd(path)
        return Path(path).absolute()



# ------------------------------------------------------------------------
# Deformable DETR
# Copyright (c) 2020 SenseTime. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 [see LICENSE for details]
# ------------------------------------------------------------------------
# Modified from DETR (https://github.com/facebookresearch/detr)
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
# ------------------------------------------------------------------------

"""
Backbone modules.
"""
try:
    import torch
    import torch.nn.functional as F
    import torchvision
    from torch import nn
    from torchvision.models._utils import IntermediateLayerGetter
    from typing import Dict, List
    from arcgis.learn._utils.coco_detection_utils import NestedTensor

    from .position_encoding import build_position_encoding
    from .swav_resnet50 import ResNet, resnet50
except:
    pass


class FrozenBatchNorm2d(torch.nn.Module):
    """
    BatchNorm2d where the batch statistics and the affine parameters are fixed.

    Copy-paste from torchvision.misc.ops with added eps before rqsrt,
    without which any other models than torchvision.models.resnet[18,34,50,101]
    produce nans.
    """

    def __init__(self, n, eps=1e-5):
        super(FrozenBatchNorm2d, self).__init__()
        self.register_buffer("weight", torch.ones(n))
        self.register_buffer("bias", torch.zeros(n))
        self.register_buffer("running_mean", torch.zeros(n))
        self.register_buffer("running_var", torch.ones(n))
        self.eps = eps

    def _load_from_state_dict(
        self,
        state_dict,
        prefix,
        local_metadata,
        strict,
        missing_keys,
        unexpected_keys,
        error_msgs,
    ):
        num_batches_tracked_key = prefix + "num_batches_tracked"
        if num_batches_tracked_key in state_dict:
            del state_dict[num_batches_tracked_key]

        super(FrozenBatchNorm2d, self)._load_from_state_dict(
            state_dict,
            prefix,
            local_metadata,
            strict,
            missing_keys,
            unexpected_keys,
            error_msgs,
        )

    def forward(self, x):
        # move reshapes to the beginning
        # to make it fuser-friendly
        w = self.weight.reshape(1, -1, 1, 1)
        b = self.bias.reshape(1, -1, 1, 1)
        rv = self.running_var.reshape(1, -1, 1, 1)
        rm = self.running_mean.reshape(1, -1, 1, 1)
        eps = self.eps
        scale = w * (rv + eps).rsqrt()
        bias = b - rm * scale
        return x * scale + bias


class BackboneBase(nn.Module):
    def __init__(
        self, backbone: nn.Module, train_backbone: bool, return_interm_layers: bool
    ):
        super().__init__()
        for name, parameter in backbone.named_parameters():
            if not train_backbone or "layer3" not in name and "layer4" not in name:
                parameter.requires_grad_(False)
        if return_interm_layers:
            # return_layers = {"layer1": "0", "layer2": "1", "layer3": "2", "layer4": "3"}
            return_layers = {"layer2": "0", "layer3": "1", "layer4": "2"}
            self.strides = [8, 16, 32]
            self.num_channels = [512, 1024, 2048]
        else:
            return_layers = {"layer4": "0"}
            self.strides = [32]
            self.num_channels = [2048]
        self.body = IntermediateLayerGetter(backbone, return_layers=return_layers)

    def forward(self, tensor_list):
        from torch import Tensor

        if isinstance(tensor_list, Tensor):
            out = self.forward_non_nested(tensor_list)
        else:
            xs = self.body(tensor_list.tensors)
            out: Dict[str, NestedTensor] = {}
            for name, x in xs.items():
                m = tensor_list.mask
                assert m is not None
                mask = F.interpolate(m[None].float(), size=x.shape[-2:]).to(torch.bool)[
                    0
                ]
                out[name] = NestedTensor(x, mask)
        # else:
        #    out = self.forward_non_nested(tensor_list)
        return out

    def forward_non_nested(self, tensors):
        xs = self.body(tensors)
        out: Dict[str, NestedTensor] = {}
        for name, x in xs.items():
            out[name] = x
        return out


class Backbone(BackboneBase):
    """ResNet backbone with frozen BatchNorm."""

    def __init__(
        self,
        name: str,
        train_backbone: bool,
        return_interm_layers: bool,
        dilation: bool,
        load_backbone: str,
    ):
        pretrained = load_backbone == "supervised"
        backbone = getattr(torchvision.models, name)(
            replace_stride_with_dilation=[False, False, dilation],
            pretrained=pretrained,
            norm_layer=FrozenBatchNorm2d,
        )
        # load the SwAV pre-training model from the url instead of supervised pre-training model
        if name == "resnet50" and load_backbone == "swav":
            checkpoint = torch.hub.load_state_dict_from_url(
                "https://dl.fbaipublicfiles.com/deepcluster/swav_800ep_pretrain.pth.tar",
                map_location="cpu",
            )
            state_dict = {k.replace("module.", ""): v for k, v in checkpoint.items()}
            backbone.load_state_dict(state_dict, strict=False)
        super().__init__(backbone, train_backbone, return_interm_layers)


class Joiner(nn.Sequential):
    def __init__(self, backbone, position_embedding):
        super().__init__(backbone, position_embedding)
        self.strides = backbone.strides
        self.num_channels = backbone.num_channels

    def forward(self, tensor_list: NestedTensor):
        xs = self[0](tensor_list)
        out: List[NestedTensor] = []
        pos = []
        for name, x in sorted(xs.items()):
            out.append(x)

        # position encoding
        for x in out:
            pos.append(self[1](x).to(x.tensors.dtype))

        return out, pos


def build_leran_backbone(backbone_name="resnet50", hidden_dim=128):
    from .position_encoding import PositionEmbeddingSine

    position_embedding = PositionEmbeddingSine(hidden_dim, normalize=True)
    backbone = Backbone(
        backbone_name,
        True,
        True,
        False,
        load_backbone="swav",
    )
    model = Joiner(backbone, position_embedding)
    return model


def build_backbone(args):
    position_embedding = build_position_encoding(args)
    train_backbone = args.lr_backbone > 0
    return_interm_layers = args.masks or (args.num_feature_levels > 1)
    backbone = Backbone(
        args.backbone,
        train_backbone,
        return_interm_layers,
        args.dilation,
        load_backbone=args.load_backbone,
    )
    model = Joiner(backbone, position_embedding)
    return model


def build_swav_backbone(args, device):
    model = resnet50(
        normalize=True,
        hidden_mlp=2048,
        output_dim=128,
    )
    for name, parameter in model.named_parameters():
        parameter.requires_grad_(False)

    checkpoint = torch.hub.load_state_dict_from_url(
        "https://dl.fbaipublicfiles.com/deepcluster/swav_800ep_pretrain.pth.tar",
        map_location="cpu",
    )
    state_dict = {k.replace("module.", ""): v for k, v in checkpoint.items()}
    missing_keys, unexpected_keys = model.load_state_dict(state_dict, strict=False)
    return model.to(device)


def build_swav_backbone_old(args, device):
    train_backbone = False
    return_interm_layers = args.masks or (args.num_feature_levels > 1)
    model = Backbone(
        args.backbone,
        train_backbone,
        return_interm_layers,
        args.dilation,
        load_backbone=args.load_backbone,
    ).to(device)

    def model_func(elem):
        return model(elem)["0"].mean(dim=(2, 3))

    return model_func



from __future__ import annotations
from dataclasses import dataclass
from typing import Any, Optional, Union
from arcgis import env
from arcgis._impl.common._mixins import PropertyMap
from arcgis.features._trace_configuration import TraceConfiguration
from arcgis._impl.common._deprecate import deprecated


########################################################################
class UtilityNetworkManager(object):
    """
    The Utility Network Service exposes analytic capabilities (tracing)
    as well as validation of network topology and management of
    subnetworks (managing sources, updating subnetworks, exporting
    subnetworks, and so on). The Utility Network Service is conceptually
    similar to the Network Analysis Service for transportation networks.

    =====================   ===========================================
    **Inputs**              **Description**
    ---------------------   -------------------------------------------
    url                     Required String. The web endpoint to the utility service.
    ---------------------   -------------------------------------------
    version                 Optional :class:`~arcgis.features._version.Version`. The `Version` class where the branch version will take place.
    ---------------------   -------------------------------------------
    gis                     Optional :class:`~arcgis.gis.GIS` . The `GIS` connection object.
    =====================   ===========================================


    """

    _con = None
    _gis = None
    _url = None
    _property = None
    _version_guid = None
    _version_name = None
    _version = None

    # ----------------------------------------------------------------------
    def __init__(self, url, version=None, gis=None):
        """Constructor"""
        if gis is None:
            gis = env.active_gis
        self._gis = gis
        self._con = gis._portal.con
        self._url = url
        if version:
            self._version = version
            self._version_name = version.properties.versionName
            if self._version_name in ["SDE.DEFAULT", "DBO.DEFAULT"]:
                # fix due to locking issue
                self._version_guid = None
            else:
                self._version_guid = version._guid
        else:
            self._version = None
            self._version_guid = None
            self._version_name = None

    # ----------------------------------------------------------------------
    def _init(self):
        """initializer"""
        try:
            res = self._con.get(self._url, {"f": "json"})
            self._property = PropertyMap(res)
        except Exception as e:
            self._property = PropertyMap({})

    # ----------------------------------------------------------------------
    def __str__(self) -> str:
        return f"< Utility Network Server @ {self._url} >"

    # ----------------------------------------------------------------------
    def __repr__(self) -> str:
        return f"< Utility Network Server @ {self._url} >"

    # ----------------------------------------------------------------------
    @property
    def properties(self):
        """returns the properties for the service"""
        if self._property is None:
            self._init()
        return self._property

    # ----------------------------------------------------------------------
    def trace(
        self,
        locations: list[dict],
        trace_type: str,
        moment: int | None = None,
        configuration: dict | TraceConfiguration | None = None,
        result_type: str | None = None,
        result_types: list[dict] | None = None,
        trace_config_global_id: str | None = None,
        out_sr: int | None = None,
        pbf: bool = False,
    ) -> dict:
        """
        A trace refers to a pre-configured algorithm that systematically
        travels a network to return results. Generalized traces allow you to
        trace across multiple types of domain networks. For example, running
        a Connected trace from your electric network through to your gas
        network. An assortment of options is provided with trace to support
        various analytic work flows. All traces use the network topology to
        read cached information about network features. This can improve
        performance of complex traces on large networks. Trace results are
        not guaranteed to accurately represent a utility network when dirty
        areas are present. The network topology must be validated to ensure
        it reflects the most recent edits or updates made to the network.

        .. note::
            The active portal account must be licensed with the ArcGIS Utility
            Network user type extention to use this operation.

        =======================    ==================================================
        **Parameter**              **Description**
        -----------------------    --------------------------------------------------
        locations                  Required list of dictionaries. The locations for
                                   starting points and barriers. An empty array must
                                   be used when performing a subnetwork trace if a
                                   subnetworkName is provided as part of the
                                   `configuration`—for example, `locations=[]`.


                                   The location is ignored by the trace if the following
                                   required properties are not defined:

                                   * `percentAlong` : required for edge features and objects.
                                   * `terminalID` : required for junction features and objects.


                                   .. code-block:: python

                                       [{
                                           "traceLocationType" : "startingPoint" | "barrier",
                                           "globalId" : <guid>,
                                           "terminalId" : <int>,   // optional
                                           "percentAlong" : <double>, // optional
                                           "isFilterBarrier" : true | false // optional Introduced at 10.8.1
                                       }]
        -----------------------    --------------------------------------------------
        trace_type                 Required string. Specifies the core algorithm that
                                   will be executed to analyze the network. Can be
                                   configured using the `configuration` parameter.

                                   Values:

                                    'connected' | 'subnetwork' | 'subnetworkController' | 'upstream' | 'downstream' | 'loops' | 'shortestPath' | 'isolation'
        -----------------------    --------------------------------------------------
        moment                     Optional Integer. Specifies the session moment. This
                                   should only be specified if you do not want to use
                                   the current moment.

                                   Example: moment = <Epoch time in milliseconds>
        -----------------------    --------------------------------------------------
        configuration              Optional dictionary or TraceConfiguration object.
                                   Specifies the collection of trace configuration
                                   properties. Depending on the `trace_type`, some
                                   properties are required.

                                   To see all configuration properties see:
                                   `Trace Configuration Properties
                                   <https://developers.arcgis.com/rest/services-reference/enterprise/trace-utility-network-server-.htm#GUID-F0C932FD-B403-4223-9B00-E44D156C7DF9/>`_
        -----------------------    --------------------------------------------------
        result_types               Optional parameter specifying hte types of results
                                   to return.

                                   .. code-block::
                                       [{
                                           "type" : "elements" | "aggregatedGeometry" | "connectivity",
                                           "includeGeometry" : true | false,
                                           "includePropagatedValues": true | false,
                                           "networkAttributeNames" :["attribute1Name","attribute2Name",...],
                                           "diagramTemplateName": <value>,
                                           "resultTypeFields":[{"networkSourceId":<int>,"fieldname":<value>},...]
                                       },...]
        -----------------------    --------------------------------------------------
        trace_config_global_id     Optional String. The global ID of a named trace configuration.
                                   When specified, this configuration is used instead of the
                                   traceConfiguration parameter. Additionally, named trace
                                   configurations are persisted with their own trace type so the
                                   trace type parameter is ignored.
        -----------------------    --------------------------------------------------
        out_sr                     Optional Integer. The output spatial reference.
        -----------------------    --------------------------------------------------
        pbf                        Optional Boolean. If True, the results are returned in
                                   the PBF format. The default is False.
        =======================    ==================================================

        :return:
            A dictionary with keys and value types of:

                | {
                |    "traceResults": {
                |        "elements": list,
                |        "diagramName": str,
                |        "globalFunctionResults": list,
                |        "kFeaturesForKNNFound": bool,
                |        "startingPointsIgnored" bool,
                |        "warnings": list
                |    }
                |    "success": bool
                | }

        """
        url = "%s/trace" % self._url
        if isinstance(configuration, TraceConfiguration):
            configuration = configuration.to_dict()
        params = {
            "f": "pbf" if pbf is True else "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "traceType": trace_type,
            "moment": moment,
            "traceLocations": locations,
            "traceConfiguration": configuration,
        }
        if trace_config_global_id:
            params["traceConfigurationGlobalId"] = trace_config_global_id
        if self._gis.version <= [7, 3]:
            params["resultType"] = result_type
        else:
            params["resultTypes"] = result_types
        if out_sr:
            params["outSR"] = out_sr
        if pbf is True:
            return self._con.post(url, params, force_bytes=True)
        else:
            return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def disable_topology(self) -> dict:
        """
        Disables the network topology for a utility network. When the
        topology is disabled, feature and association edits do not generate
        dirty areas. Analytics and diagram generation can't be performed if
        the topology is not present.

        When the topology is disabled, the following happens:

        - All current rows in the topology tables are deleted.
        - No dirty areas are generated from edits.
        - Remaining error features still exist and can be cleaned up without the overhead of dirty areas.

        To perform certain network configuration tasks, the network
        topology must be disabled.

        - This operation must be executed by the portal utility network owner.
        - The topology can be disabled in the default version or in a named version.

        :return: Dictionary indicating 'success' or 'error'

        """
        url = "%s/disableTopology" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def enable_topology(self, error_count: int = 10000) -> dict:
        """
        Enabling the network topology for a utility network is done on the
        **DEFAULT** version. Enabling is **not** supported in named versions.
        When the topology is enabled, all feature and association edits
        generate dirty areas, which are then consumed when the network
        topology is updated.

        When topology is enabled, the following happens:
        - Any existing errors are deleted.
        - The topology is updated for the full extent of the network.
        - Any newly discovered errors are added to the dirty areas sublayer.
        - The topology is marked as enabled.

        .. note::
            The active portal account must be licensed with the ArcGIS Utility
            Network user type extension to use this operation.

        ====================================     ====================================================================
        **Parameter**                             **Description**
        ------------------------------------     --------------------------------------------------------------------
        error_count                              Optional Integer. Sets the threshold when the `enable_topology` will
                                                 stop if the maximum number of errors is met. The default value is
                                                 10,000.
        ====================================     ====================================================================

        :return: Dictionary indicating 'success' or 'error'

        """
        if self._version_name.lower().find("default") == -1:
            raise Exception("Current version is not the `DEFAULT` version.")

        params = {"f": "json", "maxErrorCount": error_count}
        url = "%s/enableTopology" % self._url
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def disable_subnetwork_controller(
        self,
        network_source_id: str,
        global_id: str,
        terminal_id: str,
        out_sr: int | None = None,
    ) -> dict:
        """
        A subnetwork controller (or simply, a source or a sink) is the
        origin (or destination) of resource flow for a subpart of the
        network. Examples of subnetwork controllers are circuit breakers in
        electric networks, or town border stations in gas networks.
        Subnetwork controllers correspond to devices that have the
        Subnetwork Controller network capability set. A source is removed
        with `disable_subnetwork_controller`.

        ====================================        ====================================================================
        **Parameter**                                **Description**
        ------------------------------------        --------------------------------------------------------------------
        network_source_id                           Required String. The network source ID that the subnetwork controller
                                                    participates in.
        ------------------------------------        --------------------------------------------------------------------
        global_id                                   Required String. The global ID of the device being disabled as a
                                                    network controller.
        ------------------------------------        --------------------------------------------------------------------
        terminal_id                                 Required String. The terminal ID of the device being disabled as a
                                                    network controller.
        ------------------------------------        --------------------------------------------------------------------
        out_sr                                      Required int. The output spatial reference as a wkid.
        ====================================        ====================================================================

        """

        url = "%s/disableSubnetworkController" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "networkSourceId": network_source_id,
            "featureGlobalId": global_id,
            "terminalId": terminal_id,
        }
        if out_sr:
            params["outSR"] = out_sr
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def enable_subnetwork_controller(
        self,
        network_source_id: str,
        global_id: str,
        terminal_id: str,
        subnetwork_controller_name: str,
        tier_name: str,
        subnetwork_name: str | None = None,
        description: str | None = None,
        notes: str | None = None,
        out_sr: int | None = None,
    ) -> dict:
        """
        A subnetwork controller is the origin (or destination) of resource
        flow for a subpart of the network (e.g., a circuit breaker in
        electric networks, or a town border station in gas networks).
        Controllers correspond to Devices that have the Subnetwork
        Controller network capability set.

        ====================================        ====================================================================
        **Parameter**                                **Description**
        ------------------------------------        --------------------------------------------------------------------
        network_source_id                           Required String. The network source ID that the subnetwork controller
                                                    participates in.
        ------------------------------------        --------------------------------------------------------------------
        global_id                                   Required String. The global ID of the device being enabled as a
                                                    network controller.
        ------------------------------------        --------------------------------------------------------------------
        terminal_id                                 Required String. The terminal ID of the device being enabled as a
                                                    network controller.
        ------------------------------------        --------------------------------------------------------------------
        subnetwork_controller_name                  Required String. The name of the subnetwork controller.
        ------------------------------------        --------------------------------------------------------------------
        tier_name                                   Required String. The name of the tier.
        ------------------------------------        --------------------------------------------------------------------
        subnetwork_name                             Optional String. Specifies the name of the subnetwork.
        ------------------------------------        --------------------------------------------------------------------
        description                                 Optional String. Represents the description of the subnetwork controller.
        ------------------------------------        --------------------------------------------------------------------
        notes                                       Optional String. The notes associated with the subnetwork controller.
        ------------------------------------        --------------------------------------------------------------------
        out_sr                                      Optional Integer. The output spatial reference as a wkid (integer).
        ====================================        ====================================================================

        """

        url = "%s/enableSubnetworkController" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "networkSourceId": network_source_id,
            "featureGlobalId": global_id,
            "terminalID": terminal_id,
            "subnetworkControllerName": subnetwork_controller_name,
            "subnetworkName": subnetwork_name,
            "tierName": tier_name,
            "description": description,
            "notes": notes,
        }
        if not out_sr is None:
            params["outSR"] = out_sr
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def export_subnetwork(
        self,
        domain_name: str,
        tier_name: str,
        subnetwork_name: str,
        trace_configuration: dict | TraceConfiguration | None = None,
        export_acknowledgement: bool = False,
        result_type: str | None = None,
        result_types: list[dict] | None = None,
        moment: int | None = None,
        run_async: bool = False,
        out_sr: int | None = None,
        pbf: bool = False,
    ) -> dict:
        """
        The `export_subnetwork` operation is used to export information
        about a subnetwork into a JSON file. That information can then be
        consumed by outside systems such as outage management and asset
        tracking. The exportSubnetwork operation allows you to delete
        corresponding rows in the Subnetwork Sources table as long as the
        IsDeleted attribute is set to True. This indicates a source feeding
        the subnetwork has been removed.

        ====================================        ====================================================================
        **Parameter**                                **Description**
        ------------------------------------        --------------------------------------------------------------------
        domain_name                                 Required String. The name of the domain network of which the subnetwork
                                                    is a part.
        ------------------------------------        --------------------------------------------------------------------
        tier_name                                   Required String. The name of the tier of which the subnetwork is a part.
        ------------------------------------        --------------------------------------------------------------------
        subnetwork_name                             Required String. The name of the subnetwork.
        ------------------------------------        --------------------------------------------------------------------
        trace_configuration                         Optional Dictionary or TraceConfiguration object. Specifies the collection of trace
                                                    configuration parameters.
                                                    See: `Trace <https://developers.arcgis.com/rest/services-reference/enterprise/trace-utility-network-server-.htm#GUID-F0C932FD-B403-4223-9B00-E44D156C7DF9/>`_
        ------------------------------------        --------------------------------------------------------------------
        export_acknowledgement                      Optional Boolean. Specify whether the export is acknowledged.
        ------------------------------------        --------------------------------------------------------------------
        result_types                                Optional list of dictionaries. Specifies the type of results to return.

                                                    .. code-block:: python

                                                        [
                                                            {
                                                                "type" : "features" | "geometries" | "network" | "connectivity" | "controllers" | "associations" | "aggregatedGeometry" |
                                                                "diagram" | "elements" |  "associations",
                                                                "includeGeometry" : true | false,
                                                                "includePropagatedValues": true | false,
                                                                "includeDomainDescriptions": true | false,
                                                                "networkAttributeNames" :["attribute1Name","attribute2Name",...],
                                                                "diagramTemplateName": <value>,
                                                                "resultTypeFields":[{"networkSourceId":<int>,"fieldname":<value>},...]
                                                            },...
                                                        ]
        ------------------------------------        --------------------------------------------------------------------
        moment                                      Optional Integer. Specify the session moment if you do not want to use
                                                    the current moment.
        ------------------------------------        --------------------------------------------------------------------
        out_sr                                      Optional Integer. Optional parameter specifying the output spatial reference.
        ------------------------------------        --------------------------------------------------------------------
        pbf                                         Optional Boolean. If true, the response will be in PBF format.
        ====================================        ====================================================================

        :return:
            A dictionary with keys and value types of:

                | {
                |    "moment": int,
                |    "url": str,
                |    "subnetworkHasBeenDeleted": bool,
                |    "success": bool
                | }

        """

        url = "%s/exportSubnetwork" % self._url
        if isinstance(trace_configuration, TraceConfiguration):
            trace_configuration = trace_configuration.to_dict()
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "moment": moment,
            "domainNetworkName": domain_name,
            "tierName": tier_name,
            "subnetworkName": subnetwork_name,
            "exportAcknowledgement": export_acknowledgement,
            "traceConfiguration": trace_configuration,
            "async": run_async,
        }
        if self._gis.version <= [7, 3]:
            params["resultType"] = result_type
        else:
            params["resultTypes"] = result_types
        if out_sr:
            params["outSR"] = out_sr
        if pbf:
            params["f"] = "pbf"
            return self._con.post(url, params, force_bytes=True)
        else:
            return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def query_network_moments(
        self,
        moments_to_return: list[str] | None = None,
        moment: int | None = None,
    ) -> dict:
        """
        The `query_network_moments` operation returns the moments related
        to the network topology and operations against the topology. This
        includes when the topology was initially enabled, when it was last
        validated, when the topology was last disabled (and later enabled),
        and when the definition of the utility network was last modified.

        ====================================        ====================================================================
        **Parameter**                                **Description**
        ------------------------------------        --------------------------------------------------------------------
        moments_to_return                           Optional List of Strings. Represents the collection of validate moments to
                                                    return. Default is all.

                                                    Values:

                                                            [ "initialEnableTopology" | "fullValidateTopology" | "partialValidateTopology" | "enableTopology" | "disableTopology" | "definitionModification" | "updateIsConnected" | "indexUpdate" | "all"]
                                                    Example:

                                                        moments_to_return=["enableTopology","initialEnableTopology"]
        ------------------------------------        --------------------------------------------------------------------
        moment                                      Optional Integer. Specify the session moment if you do not want to use
                                                    the current moment.
        ====================================        ====================================================================

        :return:
            A dictionary with keys and value types of:

                | {"networkMoments": list,
                | "validateNetworkTopology": bool,
                | "success": bool}
        """
        url = "%s/queryNetworkMoments" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "momentsToReturn": moments_to_return
            if moments_to_return is not None
            else ["all"],
            "moment": moment,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    @deprecated(deprecated_in="2.1.0", removed_in=None, current_version="2.2.0")
    def query_overrides(
        self,
        attribute_ids: Optional[list[str]] = None,
        all_attributes: bool = False,
        all_connectivity: bool = False,
    ):
        """
        Network attributes support the ability to have their values
        overridden without having to edit features and validate the network
        topology (build the index). The utility network also supports the
        ability to place ephemeral connectivity (e.g., jumpers in an
        electrical network) between two devices or junctions without having
        to edit features or connectivity associations and validate the
        network topology (build the index). This operation allows the
        client to query all the overrides associated with the network
        attributes (by network attribute id). In addition, all connectivity
        overrides are returned.
        """
        url = "%s/queryOverrides" % self._url
        params = {
            "f": "json",
            "attributeIDs": attribute_ids,
            "allAttributes": all_attributes,
            "allConnectivity": all_connectivity,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def synthesize_association_geometries(
        self,
        attachment_associations: bool = False,
        connectivity_associations: bool = False,
        containment_associations: bool = False,
        count: int = 200,
        extent: dict = None,
        out_sr: int | dict[str, Any] | None = None,
        moment: int | None = None,
    ) -> dict:
        """
        The `synthesize_association_geometries` operation is used to export
        geometries representing associations that are synthesized as line
        segments corresponding to the geometries of the devices at the
        endpoints. All features associated with an association must be in
        the specified extent in order for the geometry to be synthesized.
        If only zero or one of the devices/junctions intersects the extent,
        then no geometry will be synthesized.

        ====================================        ====================================================================
        **Parameter**                                **Description**
        ------------------------------------        --------------------------------------------------------------------
        attachment_associations                     Optional Boolean. Whether to return attachment associations.
        ------------------------------------        --------------------------------------------------------------------
        connectivity_associations                   Optional Boolean. Represents whether to return connectivity associations.
        ------------------------------------        --------------------------------------------------------------------
        containment_associations                    Optional Boolean. Whether to return containment associations.
        ------------------------------------        --------------------------------------------------------------------
        count                                       Required Int. Represents the maximum number of geometries that
                                                    can be synthesized and returned in the result.
        ------------------------------------        --------------------------------------------------------------------
        extent                                      Required Dictionary. Represents the envelope of the area to
                                                    synthesize association geometries.

                                                    .. code-block:: python

                                                        {
                                                            "xmin": <minimum x-coordinate>,
                                                            "ymin": <minimum y-coordinate>,
                                                            "xmax": <maximum x-coordinate>,
                                                            "ymax": <maximum y-coordinate>,
                                                            "spatialReference": {
                                                            "wkid": <spatial reference well-known identifier>,
                                                            "latestWkid": <the current wkid value associated with the wkid>
                                                            }
                                                        }
        ------------------------------------        --------------------------------------------------------------------
        out_sr                                      Optional Dictionary. Represents the output spatial reference.
        ------------------------------------        --------------------------------------------------------------------
        moment                                      Optional Integer. Specify the session moment if you do not want to use
                                                    the current moment.
        ====================================        ====================================================================

        :return:
            A dictionary with keys and value types of:

                | {"maxGeometryCountExceeded": bool,
                | "associations": list,
                | "success": bool}
        """
        url = "%s/synthesizeAssociationGeometries" % self._url
        params = {
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "moment": moment,
            "attachmentAssociations": attachment_associations,
            "connectivityAssociations": connectivity_associations,
            "containmentAssociations": containment_associations,
            "maxGeometryCount": count,
            "extent": extent,
            "outSR": out_sr,
            "f": "json",
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def update_is_connected(self) -> dict:
        """

        Utility network features have an attribute called IsConnected that
        lets you know if a feature is connected to a source or sink, and
        therefore it could potentially be part of an existing subnetwork.
        The `update_is_connected` operation updates this attribute on
        features in the specified utility network. This operation can only
        be executed on the default version by the portal utility network
        owner.
        """
        url = "%s/updateIsConnected" % self._url
        params = {"f": "json"}
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def update_subnetwork(
        self,
        domain_name: str,
        tier_name: str,
        subnetwork_name: str | None = None,
        all_subnetwork_tier: bool = False,
        continue_on_failure: bool = False,
        trace_configuration: dict | None = None,
    ) -> dict:
        """
        A subnetwork is updated by calling the `update_subnetwork` operation.
        With this operation, one or all of the subnetworks in a single tier
        can be updated. When a subnetwork is updated, four things can occur;
        the Subnetwork Name attribute is updated for all features in the
        subnetwork, the record representing the subnetwork inside the
        SubnetLine class is refreshed, the Subnetworks table is updated and
        finally diagrams are generated or updated for the subnetwork.

        ====================================        ====================================================================
        **Parameter**                                **Description**
        ------------------------------------        --------------------------------------------------------------------
        domain_name                                 Required String. The name fo the domain network that the subnetwork
                                                    is a part of.
        ------------------------------------        --------------------------------------------------------------------
        tier_name                                   Required String. The name of the tier that the subnetwork is a part of.
        ------------------------------------        --------------------------------------------------------------------
        subnetwork_name                             Optional String. Represents the name of the subnetwork to update. If
                                                    this parameter is not specified, the `all_subnetwork_tier` parameter
                                                    should be set to `True`. Otherwise an error will occur.
        ------------------------------------        --------------------------------------------------------------------
        all_subnetwork_tier                         Optional Bool. Set to `True` when all the subnetworks in a tier
                                                    need to be updated.
        ------------------------------------        --------------------------------------------------------------------
        continue_on_failure                         Optional Bool. Continue updating subnetworks when `all_subnetwork_tier`
                                                    is `True` and a failure occurs when processing a subnetwork.
        ------------------------------------        --------------------------------------------------------------------
        trace_configuration                         Optional Dictionary. Represents the collection of trace configuration
                                                    parameters. See `trace` method to get parameters.
        ====================================        ====================================================================

        :return: Dictionary of the JSON response.

        """
        url = "%s/updateSubnetwork" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "domainNetworkName": domain_name,
            "tierName": tier_name,
            "subnetworkName": subnetwork_name,
            "allSubnetworksInTier": all_subnetwork_tier,
            "continueOnFailure": continue_on_failure,
            "traceConfiguration": trace_configuration,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def validate_topology(
        self,
        envelope: dict[str, Any],
        run_async: bool = False,
        return_edits: bool = False,
        validate_set: list[dict] | None = None,
        out_sr: int | None = None,
    ) -> dict:
        """
        Validating the network topology for a utility network maintains
        consistency between feature editing space and network topology space.
        Validating a network topology may include all or a subset of the
        dirty areas present in the network. Validation of network topology
        is supported synchronously and asynchronously.

        ====================================        ====================================================================
        **Parameter**                                **Description**
        ------------------------------------        --------------------------------------------------------------------
        envelope                                    Required Dictionary. The envelope of the area to validate.

                                                    .. code-block:: python

                                                        {
                                                            "xmin": <minimum x-coordinate>,
                                                            "ymin": <minimum y-coordinate>,
                                                            "xmax": <maximum x-coordinate>,
                                                            "ymax": <maximum y-coordinate>,
                                                            "spatialReference": {
                                                            "wkid": <spatial reference well-known identifier>,
                                                            "latestWkid": <the current wkid value associated with the wkid>
                                                            }
                                                        }
        ------------------------------------        --------------------------------------------------------------------
        run_async                                   Optional Boolean. If Turem the request is processed as an asynchronous
                                                    job. The URL is returned to check the status of a job.
        ------------------------------------        --------------------------------------------------------------------
        return_edits                                Optional Boolean. Returned results are organized in a layer-by-layer fashion.
                                                    If `return_edits` is set to True, each layer may have edited features
                                                    returned in an editedFeatures object.
                                                    The editedFeatures object returns full features including the original
                                                    features prior to delete; the original and current features for updates;
                                                    and the current rows for inserts, which may contain implicit changes
                                                    (for example, as a result of a calculation rule).

                                                    The response includes no editedFeatures and 'exceededTransferLimit = true'
                                                    if the count of edited features to return is more than the maxRecordCount.
                                                    If clients are using this parameter to maintain a cache, they should
                                                    invalidate the cache when exceededTransferLimit = true is returned.
                                                    If the server encounters an error when generating the list
                                                    of edits is the response, exceededTransferLimit = true is also returned.

                                                    Edited features are returned in the spatial reference
                                                    of the feature service as defined by the service's spatialReferenceobject
                                                    or by the spatialReference of the layer's extent object.
        ------------------------------------        --------------------------------------------------------------------
        validate_set                                Optional List of Dictionary. Introduced at Enterprise 10.9.1, it specifies
                                                    the set of features and objects to validate.

                                                    .. code-block:: python

                                                        [
                                                            {
                                                                "sourceId": <int>,
                                                                "globalIds": [<guid>]
                                                            }
                                                        ]
        ------------------------------------        --------------------------------------------------------------------
        out_sr                                      Optional integer. The output spatial reference.
        ====================================        ====================================================================

        :return: Dictionary indicating 'success' or 'error'

        """
        url = "%s/validateNetworkTopology" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "validateArea": envelope,
            "async": run_async,
            "returnEdits": return_edits,
        }
        if self._gis.version >= [9, 2]:
            params["validateSet"] = validate_set
        if out_sr:
            params["outSR"] = out_sr
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    @deprecated(deprecated_in="2.1.0", removed_in=None, current_version="2.2.0")
    def apply_overrides(
        self,
        adds: Optional[Union[list, dict[str, Any]]] = None,
        deletes: Optional[Union[list, dict[str, Any]]] = None,
    ):
        """
        Network attributes support the ability to have their values
        overridden without having to edit features and validate the network
        topology (build the index). The utility network also supports the
        ability to place ephemeral connectivity (for example, jumpers in an
        electrical network) between two devices or junctions without having
        to edit features or connectivity associations and validate the
        network topology (build the index). When specified by the client, a
        trace operation may optionally incorporate the network attribute
        and connectivity override values when the trace is run on.


        """
        url = "%s/applyOverrides" % self._url
        params = {"f": "json", "adds": adds, "deletes": deletes}
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def associations(self) -> dict:
        """
        The associations resource provides access to operations that
        allow you to query and extract useful information from the
        associations table of a utility network.

        Available starting at Enterprise 10.9.1

        :return: A dictionary with two keys

            {"associations":list, "success": bool}
        """

        if self._gis.version >= [9, 2]:
            url = "%s/associations" % self._url
            params = {"f": "json"}
            return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def query_associations(
        self,
        elements: list[dict] | None = None,
        moment: int | None = None,
        types: list[str] | None = None,
        return_deletes: bool = False,
    ) -> dict:
        """
        The query operation allows you to query the associations table
        and return association information for network features in a utility network.

        Available starting at Enterprise 10.9.1

        ====================================        ====================================================================
        **Parameter**                                **Description**
        ------------------------------------        --------------------------------------------------------------------
        elements                                    Required List of Dictionary. The feature or object elements for which
                                                    the association is querried.

                                                    .. code-block:: python

                                                        [{
                                                            "networkSourceId": <int>,
                                                            "globalId" : <guid>,
                                                            "terminalId": <int> //optional
                                                        }]
        ------------------------------------        --------------------------------------------------------------------
        moment                                      Optional Epoch time in milliseconds. Specify if you do not want to
                                                    use the current moment.
        ------------------------------------        --------------------------------------------------------------------
        types                                       Optional List of String(s). Specify the association types to be queried.

                                                    Values:

                                                        "connectivity" | "attachment" | "contianment" | "junctionEdgeFromConnectivity" | "junctionMidspanConnectivity" | "junctionEdgeToConnectivity"
        ------------------------------------        --------------------------------------------------------------------
        return_deletes                              Optional Boolean. Specify whether to return logically deleted associations.
        ====================================        ====================================================================

        """
        if self._gis.version >= [9, 2]:
            url = "%s/associations/query" % self._url
            params = {
                "f": "json",
                "gdbVersion": self._version_name,
                "elements": elements,
                "returnDeletes": return_deletes,
            }
            if elements:
                params["moment"] = moment
            if types:
                params["types"] = types

            return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def traverse_associations(
        self,
        elements: list[dict],
        moment: int | None = None,
        type: str = "unspecified",
        direction: str = "descending",
        dirty_filter: str = "none",
        error_filter: str = "none",
        stop_at_first_spatial: bool = True,
        max_depth: int | None = None,
    ) -> dict:
        """
        The `traverse_associations` operation allows you to obtain and extract useful
        information from the associations table in a utility network.

        The `type` parameter is used to provide the following predefined traversal types:

        * dirtyAreaExpansion—Returns associations and objects that have been modified and are marked as dirty. Completes a downward traversal, followed by an ascending traversal, with an exit filter on the first spatial feature in each direction.

        * firstContainers—Completes an ascending traversal on containment associations, with an exit filter on the first spatial feature.

        * spatialParents—Completes an ascending traversal on all association types, with an exit filter on the first spatial feature.

        * topContainers—Completes an ascending traversal to return associations and objects with no exit filter.

        * errorsNotModified—Completes a downward traversal to return associations in error, with an exit filter on the first spatial feature.

        * modifiedObjects—Completes a downward traversal to return associations that are dirty, with an exit filter on the first spatial feature.

        To create a custom traversal the `direction`, `dirty_filter`, `error_filter`, `stop_at_first_spatial`, and `max_depth` parameters can be used. When a traversal type is specified using the type parameter other than the default `unspecified``, these parameters are ignored.

        Available starting at Enterprise 10.9.1

        .. note::
            Associations are not traversed from spatial features to nonspatial objects
            and back to spatial features when the exit filter is placed on the first
            spatial feature.

        ====================================        ====================================================================
        **Parameter**                                **Description**
        ------------------------------------        --------------------------------------------------------------------
        elements                                    Required List of Dictionary. The feature or object elements for which
                                                    the association is queried.

                                                    .. code-block:: python

                                                        [{
                                                            "networkSourceId": <int>,
                                                            "globalId" : <guid>,
                                                            "terminalId": <int> //optional
                                                        }]
        ------------------------------------        --------------------------------------------------------------------
        moment                                      Optional Epoch time in milliseconds. Specify if you do not want to
                                                    use the current moment.
        ------------------------------------        --------------------------------------------------------------------
        type                                        Optional String. Specify the association types to be queried.

                                                    Values:

                                                        "unspecified" | "dirtyAreaExpansion" | "firstContainers" | "spatialParents" | "topContainers" | "errorsNotModified" | "modifiedObjects"
        ------------------------------------        --------------------------------------------------------------------
        direction                                   Optional String. Specify the direction of the association traversal.

                                                    Values:

                                                        "ascending" | "descending"
        ------------------------------------        --------------------------------------------------------------------
        dirty_filter                                Optional String. Specify whether to filter based on the dirty status
                                                    of the association.

                                                    .. note::
                                                        When `dirty_filter` and `error_filter` are specified together,
                                                        the filters are combined using the AND expression

                                                    Values:

                                                        "none" | "dirty" | "notDirty"
        ------------------------------------        --------------------------------------------------------------------
        error_filter                                Optional String. Specify whether to filter associations based on the
                                                    error code.

                                                    Values:

                                                       "none" | "inError" | "notInError"
        ------------------------------------        --------------------------------------------------------------------
        stop_at_first_spatial                       Optional Bool. Specify whether to stop the traversal of associations
                                                    from nonspatial objext to feature when a spatial feature is encountered.
                                                    The traversal will stop at the feature and will not traverse to the
                                                    next nonspatial object.
        ------------------------------------        --------------------------------------------------------------------
        max_depth                                   Optional Integer. Control how many hops through the association graph
                                                    are allowed in either the ascending or descending direction.
        ====================================        ====================================================================

        """
        if self._gis.version >= [9, 2]:
            url = "%s/associations/traverse" % self._url
            params = {
                "f": "json",
                "gdbVersion": self._version_name,
                "elements": elements,
                "moment": moment,
                "type": type,
                "direction": direction,
                "dirtyStatusFilter": dirty_filter,
                "errorStatusFilter": error_filter,
                "stopAtFirstSpatial": stop_at_first_spatial,
                "maxDepth": max_depth,
            }
            return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def locations(self) -> dict:
        """
        The `locations` resource provides access to an operation that allows
        you to query the locatability of a provided set of objects and
        optionally synthesize geometry to be returned.

        Introduced at Enterprise 10.9.1

        :return: "success" if able to reach locations, else "error"
        """

        if self._gis.version >= [9, 2]:
            url = "%s/locations" % self._url
            params = {"f": "json"}
            return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def query_locations(
        self,
        elements: list[dict],
        max_geom_count: int,
        moment: int | None = None,
        attachment_associations: bool = False,
        connectivity_associations: bool = False,
        containment_associations: bool = False,
        locations: bool = False,
        out_sr: int | dict | None = None,
    ) -> dict:
        """
        The query operation queries the locatability of the provided set of objects
        and optionally synthesizes geometry to be returned for each object in a
        geometry bag as a collection of points and polylines.

        ====================================        ====================================================================
        **Parameter**                                **Description**
        ------------------------------------        --------------------------------------------------------------------
        elements                                    Required List of Dictionary. The set of objects for which to get
                                                    locatability and synthesize the geometries.

                                                    .. code-block:: python

                                                        [{
                                                            "sourceId": <int>,
                                                            "globalIds" : [<guid>],
                                                        }]
        ------------------------------------        --------------------------------------------------------------------
        max_geom_count                              Required Int. The maximum number of geometries that can be synthesized
                                                    and returned in the result.
        ------------------------------------        --------------------------------------------------------------------
        moment                                      Optional Epoch time in milliseconds. Specify if you do not want to
                                                    use the current moment.
        ------------------------------------        --------------------------------------------------------------------
        attachment_associations                     Optional Boolean. Whether to synthesize the geometry representing the
                                                    structural attachment associations.
        ------------------------------------        --------------------------------------------------------------------
        conectivity_associations                    Optional Boolean. Whether to synthesize the geometry representing the
                                                    connectivity associations.
        ------------------------------------        --------------------------------------------------------------------
        containment_associations                    Optional Boolean. Whether to synthesize the geometry representing the
                                                    containment associations.
        ------------------------------------        --------------------------------------------------------------------
        locations                                   Optional Bool. Specify whether to synthesize the geometry representing
                                                    the derived location of the object. This option only affects the
                                                    results when objects are features or nonspatial objects.
        ------------------------------------        --------------------------------------------------------------------
        out_sr                                      Optional Dictionary or Integer. The output spatial reference.
        ====================================        ====================================================================

        :return:
            A dictionary with keys and value types of

                | {"exceededTransferLimit": bool,
                | "objects": list,
                | "associations": list,
                | "success": bool}
        """

        if self._gis.version >= [9, 2]:
            url = "%s/locations/query" % self._url
            params = {
                "f": "json",
                "gdbVersion": self._version_name,
                "sessionId": self._version_guid,
                "objects": elements,
                "maxGeometryCount": max_geom_count,
                "moment": moment,
                "attachmentAssociations": attachment_associations,
                "connectivityAssociations": connectivity_associations,
                "containmentAssociations": containment_associations,
                "locations": locations,
                "outSR": out_sr,
            }
            return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def trace_configurations(self) -> TraceConfigurationsManager:
        """
        The `trace_configurations` resource provides access to all trace
        configuration operations for a utility network.
        It is returned as an array of named trace configurations with the creator,
        name, and global ID for each.

        :return: An instance of TraceConfigurationsManager Class
        """

        if self._gis.version >= [9, 2]:
            url = "%s/traceConfigurations" % self._url
            return TraceConfigurationsManager(
                url,
                version=self._version,
                gis=self._gis,
                service_url=self._url,
            )

    # ----------------------------------------------------------------------


class TraceConfigurationsManager(object):
    """
    The traceConfigurations resource provides access to all trace configuration
    operations for a network service. It is returned as an array of named trace
    configurations with the creator, name, and global ID for each.


    The TraceConfigurationsManager allows methods to be done on a trace configuration.
    """

    _con = None
    _gis = None
    _url = None
    _property = None
    _version_guid = None
    _version_name = None
    _service_url = None

    # ----------------------------------------------------------------------
    def __init__(self, url, version=None, gis=None, service_url=None):
        """Constructor"""
        if gis is None:
            gis = env.active_gis
        self._gis = gis
        self._con = gis._portal.con
        self._url = url
        self._service_url = service_url
        if version:
            self._version = version
            self._version_guid = version._guid
            self._version_name = version.properties.versionName
        else:
            self._version = None
            self._version_guid = None
            self._version_name = None

    # ----------------------------------------------------------------------
    def __str__(self) -> str:
        return f"< Trace Configuration Manager @ {self._url} >"

    # ----------------------------------------------------------------------
    def __repr__(self) -> str:
        return f"< Trace Configuration Manager @ {self._url} >"

    # ----------------------------------------------------------------------
    def list(self) -> dict:
        """
        List of all trace configurations in a service.

        :return:
            A dictionary with two keys: {"traceConfigurations": list, "success": bool}
        """
        return self._con.post(self._url, {"f": "json"})

    # ----------------------------------------------------------------------
    def query(
        self,
        global_ids: list[str] | None = None,
        creators: list[str] | None = None,
        tags: list[str] | None = None,
        names: list[str] | None = None,
        as_trace_configuration_class: bool = False,
    ) -> dict:
        """
        The query operation returns all properties from one or more
        named trace configurations in a utility network.

        ============================        ===========================================
        **Parameter**                        **Description**
        ----------------------------        -------------------------------------------
        global_ids                          Optional list of strings. Specify the global
                                            IDs of the named trace configs to be queried.
        ----------------------------        -------------------------------------------
        creators                            Optional list of strings. The creators of
                                            the named trace configurations to be queried.
        ----------------------------        -------------------------------------------
        tags                                Optional list of strings. The user tags of
                                            the named trace configurations to be queried.
        ----------------------------        -------------------------------------------
        names                               Optional list of strings. The names of the
                                            named trace configurations to be queried.
        ----------------------------        -------------------------------------------
        as_trace_configuration_class        Optional boolean. If True the list for
                                            "traceCongifurations" in return will be a list
                                            of TraceConfiguration class instances. If False,
                                            the list will be a list of dictionaries of trace
                                            configuration instances. The default is False.
        ============================        ===========================================

        :return:
            A dictionary with two keys: {"traceConfigurations": list, "success": bool}
        """
        if self._gis.version >= [9, 2]:
            url = "%s/query" % self._url
            params = {
                "f": "json",
                "globalIds": global_ids,
                "creators": creators,
                "tags": tags,
                "names": names,
            }
            res = self._con.post(url, params)
            if as_trace_configuration_class:
                trace_configs = []
                for trace in res["traceConfigurations"]:
                    trace = TraceConfiguration.from_config(trace)
                    trace_configs.append(trace)
                res["traceConfigurations"] = trace_configs
            return res

    # ----------------------------------------------------------------------
    def delete(self, global_ids: list[str]) -> dict:
        """
        The delete operation provides the ability to delete one or more named
        trace configurations in a utility network. A named trace configuration
        can only be deleted by an administrator or its creator.

        :return: A dictionary with key "success" indicating True or False.
        """
        if self._gis.version >= [9, 2]:
            url = "%s/delete" % self._url
            params = {
                "f": "json",
                "globalIds": global_ids,
            }
            return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def create(
        self,
        name: str,
        trace_type: str,
        trace_config: dict | TraceConfiguration,
        description: str | None = None,
        result_types: list[dict] | None = None,
        tags: list[str] | None = None,
    ) -> dict:
        """
        The create operation on the traceConfigurations resource provides the
        ability to create a single named trace configuration. Named trace
        configurations store the properties of a complex trace in a utility
        network and can be shared through a map service consumed by a web map
        or field app. Multiple parameters and properties are provided with
        the create operation that support the analytic workflows associated
        with the trace operation.

        If your trace configuration already exists, use the query method to find it.

        ======================      ===============================================
        **Parameter**                **Description**
        ----------------------      -----------------------------------------------
        name                        Required String. The altered name of the trace
                                    configuration.
        ----------------------      -----------------------------------------------
        trace_type                  Required String. Specify the core algorithm that
                                    will be used to analyze the network. Trace types
                                    can be configured using the `trace_config` parameter.

                                    Values:

                                        "connected" | "subnetwork" | "upstream" | "subnetworkController" | "downstream" | "loops" | "shortenPath" | "isolation"
        ----------------------      -----------------------------------------------
        trace_config                Required Dictionary or TraceConfiguration object.
                                    Specify the collection of
                                    altered trace configuration properties.

                                    See: `Properties <https://developers.arcgis.com/rest/services-reference/enterprise/trace-utility-network-server-.htm#GUID-F0C932FD-B403-4223-9B00-E44D156C7DF9/>`_
        ----------------------      -----------------------------------------------
        description                 Optional String. Specify the altered description
                                    of the trace configuration.
        ----------------------      -----------------------------------------------
        result_types                Optional List of Dictionary. Specify the altered
                                    types of results to return.

                                    .. code-block:: python

                                        [{
                                            "type" : "elements" | "aggregatedGeometry",
                                            "includeGeometry" : true | false,
                                            "includePropagatedValues": true | false,
                                            "networkAttributeNames" :["attribute1Name","attribute2Name",...],
                                            "diagramTemplateName": <value>,
                                            "resultTypeFields":[{"networkSourceId":<int>,"fieldname":<value>},...]
                                        },...]

        ----------------------      -----------------------------------------------
        tags                        Optional List of String(s). Specify the altered
                                    user-provided tags.
        ======================      ===============================================

        :return: A dictionary with key "success" indicating True or False.
        """
        if isinstance(trace_config, TraceConfiguration):
            trace_config = trace_config.to_dict()
        if self._gis.version >= [9, 2]:
            url = "%s/create" % self._url
            params = {
                "f": "json",
                "gdbVersion": self._version_name,
                "name": name,
                "description": description,
                "traceType": trace_type,
                "traceConfiguration": trace_config,
                "resultTypes": result_types,
                "tags": tags,
            }
            return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def alter(
        self,
        global_id: str,
        name: str | None = None,
        description: str | None = None,
        trace_type: str = "connected",
        trace_config: dict | TraceConfiguration | None = None,
        result_types: list[dict] | None = None,
        tags: list[str] | None = None,
    ) -> dict:
        """
        The alter operation provides the ability to alter a single named
        trace configuration. A named trace configuration can only be altered
        by an administrator or the creator of the configuration.
        For example, you can update an existing trace configuration to
        accommodate changes in the network or address incorrectly set parameters
        without the need to delete and re-create a trace configuration.
        This enables existing map services to continue use of the named trace
        configuration without requiring the map to be republished.

        ======================      ===============================================
        **Parameter**                **Description**
        ----------------------      -----------------------------------------------
        global_id                   Required String. Specifying the global ID of
                                    the named trace configuration to alter.
        ----------------------      -----------------------------------------------
        name                        Optional String. The altered name of the trace
                                    configuration.
        ----------------------      -----------------------------------------------
        description                 Optional String. Specify the altered description
                                    of the trace configuration.
        ----------------------      -----------------------------------------------
        trace_type                  Optional String. Specify the core algorithm that
                                    will be used to analyze the network. Trace types
                                    can be configured using the `trace_config` parameter.

                                    Values:

                                            "connected" | "subnetwork" | "upstream" | "subnetworkController" | "downstream" | "loops" | "shortenPath" | "isolation"
        ----------------------      -----------------------------------------------
        trace_config                Optional Dictionary or instance of TraceConfiguration
                                    class. Specify the collection of
                                    altered trace configuration properties.

                                    See: `Properties <https://developers.arcgis.com/rest/services-reference/enterprise/trace-utility-network-server-.htm#GUID-F0C932FD-B403-4223-9B00-E44D156C7DF9/>`_
        ----------------------      -----------------------------------------------
        result_types                Optional List of Dictionary. Specify the altered
                                    types of results to return.

                                    .. code-block:: python

                                        [{
                                            "type" : "elements" | "aggregatedGeometry",
                                            "includeGeometry" : true | false,
                                            "includePropagatedValues": true | false,
                                            "networkAttributeNames" :["attribute1Name","attribute2Name",...],
                                            "diagramTemplateName": <value>,
                                            "resultTypeFields":[{"networkSourceId":<int>,"fieldname":<value>},...]
                                        },...]

        ----------------------      -----------------------------------------------
        tags                        Optional List of String(s). Specify the altered
                                    user-provided tags.
        ======================      ===============================================

        :return: A dictionary with key "success" indicating True or False.
        """

        if self._gis.version >= [9, 2]:
            url = "%s/alter" % self._url
            if isinstance(trace_config, TraceConfiguration):
                trace_config = trace_config.to_dict()
            params = {
                "f": "json",
                "globalId": global_id,
                "name": name,
                "description": description,
                "traceType": trace_type,
                "traceConfiguration": trace_config,
                "resultTypes": result_types,
                "tags": tags,
            }
            return self._con.post(url, params)

    # ----------------------------------------------------------------------



import traceback
import warnings

warnings.filterwarnings("ignore", module="transformers")
HAS_TRANSFORMER = True

try:
    from pathlib import Path
    import json
    import os
    import torch
    from transformers import pipeline, logging

    logging.get_logger("filelock").setLevel(logging.ERROR)
    from .._utils.common import _get_device_id, _get_emd_path
    from fastprogress.fastprogress import progress_bar
except Exception as e:
    transformer_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_TRANSFORMER = False


class InferenceOnlyModel:
    def __init__(self, backbone=None, task=None, **kwargs):
        self._task = task
        self.model = None
        self._backbone = backbone

        if "working_dir" in kwargs:
            self.working_dir = kwargs.get("working_dir")
        else:
            self.working_dir = Path.cwd()
        self._pretrained_path = kwargs.get("pretrained_path")
        self._device = _get_device_id()
        self.logger = logging.get_logger()
        self.logger.setLevel(logging.ERROR)
        self._load_model()

    def _load_model(self):
        try:
            if self._pretrained_path:
                self.model = pipeline(
                    self._task, model=self._pretrained_path, device=self._device
                )
            else:
                self.model = pipeline(
                    self._task, model=self._backbone, device=self._device
                )
        except Exception as e:
            error_message = (
                f"`{self._backbone}` is not valid backbone name for {self._task} task.\n"
                f"For selecting backbone name for {self._task} task, kindly visit:- "
                f"https://huggingface.co/models?pipeline_tag={self._task} "
            )
            raise Exception(error_message)

    def _create_emd(self, name_or_path):
        emd_template = {}
        emd_template.update({"ModelName": self.__class__.__name__})
        emd_template.update({"architectures": self.model.model.config.architectures})
        path = Path(name_or_path)
        name = path.parts[-1]
        with open(os.path.join(path, f"{name}.emd"), "w") as f:
            f.write(json.dumps(emd_template))

    def save(self, name_or_path):
        """
        Saves the translator model files on a specified path on the local disk.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Path to save
                                model files on the local disk.
        =====================   ===========================================

        :return: Absolute path for the saved model
        """

        if "\\" in name_or_path or "/" in name_or_path:
            path = name_or_path
        else:
            path = os.path.join(self.working_dir, "models", name_or_path)
        self.model.save_pretrained(path)
        self._create_emd(path)
        return Path(path).absolute()

    @classmethod
    def from_model(cls, emd_path, **kwargs):
        """
        Creates an :class:`~arcgis.learn.text.SequenceToSequence` model object from an
        Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to
                                Esri Model Definition(EMD) file or the folder
                                with saved model files.
        =====================   ===========================================

        :return: :class:`~arcgis.learn.text.SequenceToSequence` Object
        """
        emd_path = _get_emd_path(emd_path)
        with open(emd_path) as f:
            emd_json = json.loads(f.read())
        model_path = Path(emd_path).parent
        cls_object = cls(pretrained_path=str(model_path))
        return cls_object



from ._codetemplate import image_translation_prf
import json
import traceback
from .._data import _raise_fastai_import_error
from ._arcgis_model import ArcGISModel, _EmptyData
import logging

logger = logging.getLogger()
try:
    from ._cyclegan_utils import (
        CycleGanLoss,
        CycleGANTrainer,
        optim,
        compute_fid_metric,
    )
    from ._cyclegan_utils import CycleGAN as CycleGAN_model
    from .._utils.cyclegan import (
        ImageTuple,
        ImageTupleList,
        ImageTupleListMS,
        show_results,
    )
    from .._utils.common import (
        get_multispectral_data_params_from_emd,
        _get_emd_path,
        ArcGISMSImage,
    )
    from torchvision import transforms
    from pathlib import Path
    from fastai.vision import DatasetType, Learner, partial, open_image, Image
    import torch
    from .._utils.env import is_arcgispronotebook

    HAS_FASTAI = True
except Exception as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_FASTAI = False


class CycleGAN(ArcGISModel):

    """
    Creates a model object which generates images of type A from type B or type B from type A.

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            :meth:`~arcgis.learn.prepare_data`  function.
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained model is
                            saved.
    ---------------------   -------------------------------------------
    gen_blocks              Optional integer. Number of ResNet blocks to use
                            in generator.
    ---------------------   -------------------------------------------
    lsgan                   Optional boolean. If True, it will use Mean Squared Error
                            else it will use Binary Cross Entropy.
    =====================   ===========================================

    :return: :class:`~arcgis.learn.CycleGAN` Object
    """

    def __init__(
        self, data, pretrained_path=None, gen_blocks=9, lsgan=True, *args, **kwargs
    ):
        super().__init__(data)
        self._check_dataset_support(data)
        self._gen_blocks = gen_blocks
        self._lsgan = lsgan
        cycle_gan = CycleGAN_model(
            self._data.n_channel,
            self._data.n_channel,
            gen_blocks=gen_blocks,
            lsgan=lsgan,
        )
        self.learn = Learner(
            data,
            cycle_gan,
            loss_func=CycleGanLoss(cycle_gan),
            opt_func=partial(optim.Adam, betas=(0.5, 0.99)),
            callback_fns=[CycleGANTrainer],
        )
        self.learn.model = self.learn.model.to(self._device)
        self._slice_lr = False
        if pretrained_path is not None:
            self.load(pretrained_path)
        self._code = image_translation_prf

        def __str__(self):
            return self.__repr__()

        def __repr__(self):
            return "<%s>" % (type(self).__name__)

    @staticmethod
    def _available_metrics():
        return ["valid_loss"]

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a :class:`~arcgis.learn.CycleGAN` object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from :meth:`~arcgis.learn.prepare_data`  function or None for
                                inferencing.
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        =====================   ===========================================

        :return: :class:`~arcgis.learn.CycleGAN` Object
        """

        if not HAS_FASTAI:
            _raise_fastai_import_error(import_exception=import_exception)

        emd_path = _get_emd_path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd["ModelFile"])

        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        model_params = emd["ModelParameters"]
        resize_to = emd.get("resize_to")
        chip_size = emd["ImageHeight"]
        is_multispec = emd.get("IsMultispectral")
        if data is None:
            data = _EmptyData(
                path=emd_path.parent, loss_func=None, c=2, chip_size=resize_to
            )
            data = get_multispectral_data_params_from_emd(data, emd)
            normalization_stats_b = dict(emd.get("NormalizationStats_b"))
            for _stat in normalization_stats_b:
                if normalization_stats_b[_stat] is not None:
                    normalization_stats_b[_stat] = torch.tensor(
                        normalization_stats_b[_stat]
                    )
                setattr(data, ("_" + _stat), normalization_stats_b[_stat])

            data.n_channel = emd.get("n_intput_channel", None)
            if data.n_channel == None:
                data.n_channel = emd.get("n_channel", None)
            data._is_empty = True
            data.emd_path = emd_path
            data.emd = emd

        data.resize_to = resize_to

        return cls(data, **model_params, pretrained_path=str(model_file))

    @property
    def _model_metrics(self):
        return self.compute_metrics()

    def _get_emd_params(self, save_inference_file):
        _emd_template = {}
        _emd_template["Framework"] = "arcgis.learn.models._inferencing"
        _emd_template["ModelConfiguration"] = "_cyclegan"
        if save_inference_file:
            _emd_template["InferenceFunction"] = "ArcGISImageTranslation.py"
        else:
            _emd_template[
                "InferenceFunction"
            ] = "[Functions]System\\DeepLearning\\ArcGISLearn\\ArcGISImageTranslation.py"
        _emd_template["ModelType"] = "CycleGAN"
        _emd_template["n_intput_channel"] = self._data.n_channel
        _emd_template["SupportsVariableTileSize"] = True
        # if self._data._is_multispectral:
        _emd_template["NormalizationStats_b"] = {
            "band_min_values": self._data._band_min_values_b,
            "band_max_values": self._data._band_max_values_b,
            "band_mean_values": self._data._band_mean_values_b,
            "band_std_values": self._data._band_std_values_b,
            "scaled_min_values": self._data._scaled_min_values_b,
            "scaled_max_values": self._data._scaled_max_values_b,
            "scaled_mean_values": self._data._scaled_mean_values_b,
            "scaled_std_values": self._data._scaled_std_values_b,
        }
        for _stat in _emd_template["NormalizationStats_b"]:
            if _emd_template["NormalizationStats_b"][_stat] is not None:
                _emd_template["NormalizationStats_b"][_stat] = _emd_template[
                    "NormalizationStats_b"
                ][_stat].tolist()
        _emd_template["n_channel"] = len(
            _emd_template["NormalizationStats_b"]["band_min_values"]
        )
        model_params = {
            "gen_blocks": self._gen_blocks,
            "lsgan": self._lsgan,
        }
        _emd_template["ModelParameters"] = model_params
        return _emd_template

    def show_results(self, rows=5, **kwargs):
        """
        Displays the results of a trained model on a part of the validation set.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional int. Number of rows of results
                                to be displayed.
        =====================   ===========================================

        **kwargs**

        =====================   ===========================================
        rgb_bands               Optional list of integers (band numbers)
                                to be considered for rgb visualization.
        =====================   ===========================================

        """
        if rows > len(self._data.valid_ds):
            rows = len(self._data.valid_ds)
        show_results(self, rows, **kwargs)
        if is_arcgispronotebook():
            from matplotlib import pyplot as plt

            plt.show()

    def predict(self, img_path, convert_to):
        """
        Predicts and display the image.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        img_path                Required path of an image.
        ---------------------   -------------------------------------------
        convert_to              'A' if we want to generate image of type 'A'
                                from type 'B' or 'B' if we want to generate
                                image of type 'B' from type 'A' where A and
                                B are the domain specifications that were
                                used while training.
        =====================   ===========================================

        """
        import numpy as np

        self.learn.model.arcgis_results = True
        img_path = Path(img_path)
        n_band = self._data.n_channel
        if self._data._is_multispectral:
            raw_img = ArcGISMSImage.open(img_path)
            if n_band > raw_img.shape[0]:
                cont = []
                last_tile = np.expand_dims(raw_img.data[raw_img.shape[0] - 1, :, :], 0)
                res = abs(n_band - raw_img.shape[0])
                for i in range(res):
                    raw_img = Image(
                        torch.tensor(np.concatenate((raw_img.data, last_tile), axis=0))
                    )
        else:
            raw_img = open_image(img_path)
        raw_img_tuple = ImageTuple(raw_img, raw_img)
        pred_tuple = self.learn.predict(raw_img_tuple)
        if convert_to == "A" or convert_to == "a":
            pred_img = pred_tuple[1][0] / 2 + 0.5
        elif convert_to == "B" or convert_to == "b":
            pred_img = pred_tuple[1][1] / 2 + 0.5

        pred_img = ArcGISMSImage(pred_img)
        pred_img = pred_img.show()
        self.learn.model.arcgis_results = False
        return pred_img

    def compute_metrics(self):
        """
        Computes Frechet Inception Distance (FID) on validation set.
        """
        fid_a = "None"
        fid_b = "None"

        if self._data._imagery_type_a == "ms" and self._data._imagery_type_b == "ms":
            print("FID metric is not supported for multispectral imagery type")
        else:
            if self._data._imagery_type_a == "RGB" and self._data.n_channel == 3:
                fid_a = "{0:1.4e}".format(compute_fid_metric(self, self._data, "a"))
            if self._data._imagery_type_b == "RGB" and self._data.n_channel == 3:
                fid_b = "{0:1.4e}".format(compute_fid_metric(self, self._data, "b"))

        return {"FID_A": fid_a, "FID_B": fid_b}

    @property
    def supported_datasets(self):
        """Supported dataset types for this model."""
        return CycleGAN._supported_datasets()

    @staticmethod
    def _supported_datasets():
        return ["CycleGAN"]
